from sirius_sdk.errors.exceptions import SiriusValidationError
from sirius_sdk.messaging.message import Message
from sirius_sdk.messaging.fields import *


ID = '@id'
TYPE = '@type'
THREAD_DECORATOR = '~thread'
THREAD_ID = 'thid'
PARENT_THREAD_ID = 'pthid'
SENDER_ORDER = 'sender_order'
RECEIVED_ORDERS = 'received_orders'
THREADING_ERROR = 'threading_error'
TIMING_ERROR = 'timing_error'

TIMING_DECORATOR = '~timing'
IN_TIME = 'in_time'
OUT_TIME = 'out_time'
STALE_TIME = 'stale_time'
EXPIRES_TIME = 'expires_time'
DELAY_MILLI = 'delay_milli'
WAIT_UNTIL_TIME = 'wait_until_time'


def check_for_attributes(partial: dict, expected_attributes: Iterable):
    for attribute in expected_attributes:
        if isinstance(attribute, tuple):
            if attribute[0] not in partial:
                raise SiriusValidationError('Attribute "{}" is missing from message: \n{}'.format(attribute[0], partial))
            if partial[attribute[0]] != attribute[1]:
                raise SiriusValidationError('Message.{}: {} != {}'.format(attribute[0], partial[attribute[0]], attribute[1]))
        else:
            if attribute not in partial:
                raise SiriusValidationError('Attribute "{}" is missing from message: \n{}'.format(attribute, partial))


def validate_common_blocks(partial: dict):
    """Validate blocks of message like threading, timing, etc
    """
    _validate_thread_block(partial)
    _validate_timing_block(partial)


def _validate_thread_block(partial: dict):
    if THREAD_DECORATOR in partial:
        thread = partial[THREAD_DECORATOR]
        check_for_attributes(thread, [THREAD_ID])

        thread_id = thread[THREAD_ID]
        if partial.get(ID) and thread_id == partial[ID]:
            raise SiriusValidationError('Thread id {} cannot be equal to outer id {}'.format(thread_id, partial[ID]))
        if thread.get(PARENT_THREAD_ID) and thread[PARENT_THREAD_ID] in (thread_id, partial[ID]):
            raise SiriusValidationError('Parent thread id {} must be different than thread id and outer id'.format(
                thread[PARENT_THREAD_ID]))

        if thread.get(SENDER_ORDER):
            non_neg_num = NonNegativeNumberField()
            err = non_neg_num.validate(thread[SENDER_ORDER])
            if not err:
                if RECEIVED_ORDERS in thread and thread[RECEIVED_ORDERS]:
                    recv_ords = thread[RECEIVED_ORDERS]
                    err = MapField(DIDField(), non_neg_num).validate(recv_ords)
            if err:
                raise ValueError(err)


def _validate_timing_block(partial: dict):
    if TIMING_DECORATOR in partial:
        timing = partial[TIMING_DECORATOR]
        non_neg_num = NonNegativeNumberField()
        iso_data = ISODatetimeStringField()
        expected_iso_fields = [IN_TIME, OUT_TIME, STALE_TIME, EXPIRES_TIME, WAIT_UNTIL_TIME]
        for f in expected_iso_fields:
            if f in timing:
                err = iso_data.validate(timing[f])
                if err:
                    raise SiriusValidationError(err)
        if DELAY_MILLI in timing:
            err = non_neg_num.validate(timing[DELAY_MILLI])
            if err:
                raise SiriusValidationError(err)

        # In time cannot be greater than out time
        if IN_TIME in timing and OUT_TIME in timing:
            t_in = iso_data.parse_func(timing[IN_TIME])
            t_out = iso_data.parse_func(timing[OUT_TIME])

            if t_in > t_out:
                raise SiriusValidationError('{} cannot be greater than {}'.format(IN_TIME, OUT_TIME))

        # Stale time cannot be greater than expires time
        if STALE_TIME in timing and EXPIRES_TIME in timing:
            t_stale = iso_data.parse_func(timing[STALE_TIME])
            t_exp = iso_data.parse_func(timing[EXPIRES_TIME])

            if t_stale > t_exp:
                raise SiriusValidationError('{} cannot be greater than {}'.format(STALE_TIME, EXPIRES_TIME))



from abc import abstractmethod
import contextlib
from typing import Union
from datetime import datetime, timedelta

import sirius_sdk
from sirius_sdk.agent.pairwise import Pairwise
from sirius_sdk.hub import CoProtocolP2P
from sirius_sdk.agent.ledger import Ledger
from sirius_sdk.agent.aries_rfc.utils import utc_to_str
from sirius_sdk.agent.wallet.abstract.cache import CacheOptions
from sirius_sdk.base import AbstractStateMachine
from sirius_sdk.agent.aries_rfc.feature_0015_acks import Ack, Status
from sirius_sdk.agent.aries_rfc.feature_0037_present_proof.messages import *


PROPOSE_NOT_ACCEPTED = "propose_not_accepted"
RESPONSE_NOT_ACCEPTED = "response_not_accepted"
RESPONSE_PROCESSING_ERROR = "response_processing_error"
REQUEST_NOT_ACCEPTED = "request_not_accepted"
RESPONSE_FOR_UNKNOWN_REQUEST = "response_for_unknown_request"
REQUEST_PROCESSING_ERROR = 'request_processing_error'
VERIFY_ERROR = 'verify_error'


class BaseVerifyStateMachine(AbstractStateMachine):

    def __init__(self, time_to_live: int = 60, logger=None, *args, **kwargs):
        super().__init__(time_to_live=time_to_live, logger=logger, *args, **kwargs)
        self._problem_report: Optional[PresentProofProblemReport] = None
        self.__time_to_live = time_to_live
        self.__coprotocol: Optional[CoProtocolP2P] = None

    @property
    def time_to_live(self) -> Optional[int]:
        return self.__time_to_live

    @property
    def problem_report(self) -> PresentProofProblemReport:
        return self._problem_report

    @contextlib.asynccontextmanager
    async def coprotocol(self, pairwise: Pairwise):
        self.__coprotocol = CoProtocolP2P(
            pairwise=pairwise,
            protocols=[BasePresentProofMessage.PROTOCOL, Ack.PROTOCOL],
            time_to_live=self.time_to_live
        )
        self._register_for_aborting(self.__coprotocol)
        try:
            try:
                yield
            except OperationAbortedManually:
                await self.log(progress=100, message='Aborted')
                raise StateMachineAborted('Aborted by User')
        finally:
            self._unregister_for_aborting(self.__coprotocol)

    async def switch(self, request: BasePresentProofMessage) -> Union[BasePresentProofMessage, Ack]:
        ok, resp = await self.__coprotocol.switch(request)
        if ok:
            if isinstance(resp, BasePresentProofMessage) or isinstance(resp, Ack):
                try:
                    resp.validate()
                except SiriusValidationError as e:
                    raise StateMachineTerminatedWithError(
                        RESPONSE_PROCESSING_ERROR if self._is_leader() else REQUEST_PROCESSING_ERROR,
                        e.message
                    )
                return resp
            elif isinstance(resp, PresentProofProblemReport):
                raise StateMachineTerminatedWithError(resp.problem_code, resp.explain, notify=False)
            else:
                raise StateMachineTerminatedWithError(
                    RESPONSE_PROCESSING_ERROR if self._is_leader() else REQUEST_PROCESSING_ERROR,
                    'Unexpected response @type: %s' % str(resp.type)
                )
        else:
            raise StateMachineTerminatedWithError(
                RESPONSE_PROCESSING_ERROR if self._is_leader() else REQUEST_PROCESSING_ERROR,
                'Response awaiting terminated by timeout'
            )

    async def send(self, msg: Union[BasePresentProofMessage, Ack, PresentProofProblemReport]):
        await self.__coprotocol.send(msg)

    @abstractmethod
    def _is_leader(self) -> bool:
        raise NotImplemented


class Verifier(BaseVerifyStateMachine):
    """Implementation of Verifier role for present-proof protocol

    See details: https://github.com/hyperledger/aries-rfcs/tree/master/features/0037-present-proof
    """

    def __init__(self, prover: Pairwise, ledger: Ledger, time_to_live: int = 60, logger=None, *args, **kwargs):
        """
        :param prover: Prover described as pairwise instance.
          (Assumed pairwise was established earlier: statically or via connection-protocol)
        :param ledger: network (DKMS) name that is used to verify credentials presented by prover
          (Assumed Ledger was configured earlier - pool-genesis file was uploaded and name was set)
        """

        super().__init__(time_to_live=time_to_live, logger=logger, *args, **kwargs)
        self.__prover = prover
        self.__pool_name = ledger.name
        self.__requested_proof = None

    @property
    def requested_proof(self) -> Optional[dict]:
        return self.__requested_proof

    async def verify(
            self, proof_request: dict, translation: List[AttribTranslation] = None,
            comment: str = None, locale: str = BasePresentProofMessage.DEF_LOCALE, proto_version: str = None
    ) -> bool:
        """
        :param proof_request: Hyperledger Indy compatible proof-request
        :param translation: human readable attributes translations
        :param comment: human readable comment from Verifier to Prover
        :param locale: locale, for example "en" or "ru"
        :param proto_version: 0037 protocol version, for example 1.0 or 1.1
        """
        async with self.coprotocol(pairwise=self.__prover):
            try:
                # Step-1: Send proof request
                expires_time = datetime.utcnow() + timedelta(seconds=self.time_to_live)
                request_msg = RequestPresentationMessage(
                    proof_request=proof_request,
                    translation=translation,
                    comment=comment,
                    locale=locale,
                    expires_time=utc_to_str(expires_time),
                    version=proto_version
                )
                request_msg.please_ack = True
                await self.log(progress=30, message='Send request', payload=dict(request_msg))

                # Switch to await participant action
                presentation = await self.switch(request_msg)
                if not isinstance(presentation, PresentationMessage):
                    raise StateMachineTerminatedWithError(
                        RESPONSE_NOT_ACCEPTED, 'Unexpected @type: %s' % str(presentation.type)
                    )
                await self.log(progress=60, message='Presentation received')

                # Step-2 Verify
                identifiers = presentation.proof.get('identifiers', [])
                schemas = {}
                credential_defs = {}
                rev_reg_defs = {}
                rev_regs = {}
                opts = CacheOptions()
                for identifier in identifiers:
                    schema_id = identifier['schema_id']
                    cred_def_id = identifier['cred_def_id']
                    rev_reg_id = identifier['rev_reg_id']
                    if schema_id and schema_id not in schemas:
                        schemas[schema_id] = await sirius_sdk.Cache.get_schema(
                            self.__pool_name, self.__prover.me.did, schema_id, opts
                        )
                    if cred_def_id and cred_def_id not in credential_defs:
                        credential_defs[cred_def_id] = await sirius_sdk.Cache.get_cred_def(
                            self.__pool_name, self.__prover.me.did, cred_def_id, opts
                        )
                success = await sirius_sdk.AnonCreds.verifier_verify_proof(
                    proof_request=proof_request,
                    proof=presentation.proof,
                    schemas=schemas,
                    credential_defs=credential_defs,
                    rev_reg_defs=rev_reg_defs,
                    rev_regs=rev_regs
                )
                if success:
                    self.__requested_proof = presentation.proof['requested_proof']
                    ack = Ack(
                        thread_id=presentation.ack_message_id if presentation.please_ack else presentation.id,
                        status=Status.OK
                    )
                    await self.log(progress=100, message='Verifying terminated successfully')
                    await self.send(ack)
                    return True
                else:
                    await self.log(progress=100, message='Verifying terminated with ERROR')
                    raise StateMachineTerminatedWithError(VERIFY_ERROR, 'Verifying return false')
            except StateMachineTerminatedWithError as e:
                self._problem_report = PresentProofProblemReport(
                    e.problem_code, e.explain
                )
                await self.log(
                    progress=100, message=f'Terminated with error',
                    problem_code=e.problem_code, explain=e.explain
                )
                if e.notify:
                    await self.send(self._problem_report)
                return False

    def _is_leader(self) -> bool:
        return True


class Prover(BaseVerifyStateMachine):
    """Implementation of Prover role for present-proof protocol

    See details: https://github.com/hyperledger/aries-rfcs/tree/master/features/0037-present-proof
    """

    def __init__(self, verifier: Pairwise, ledger: Ledger, time_to_live: int = 60, logger=None, *args, **kwargs):
        """
        :param verifier: Verifier described as pairwise instance.
          (Assumed pairwise was established earlier: statically or via connection-protocol)
        :param ledger: network (DKMS) name that is used to verify credentials presented by prover
          (Assumed Ledger was configured earlier - pool-genesis file was uploaded and name was set)
        """

        super().__init__(time_to_live=time_to_live, logger=logger, *args, **kwargs)
        self.__verifier = verifier
        self.__pool_name = ledger.name

    async def prove(self, request: RequestPresentationMessage, master_secret_id: str) -> bool:
        """
        :param request: Verifier request
        :param master_secret_id: prover secret id
        """
        async with self.coprotocol(pairwise=self.__verifier):
            try:
                # Step-1: Process proof-request
                await self.log(progress=10, message='Received proof request', payload=dict(request))
                try:
                    request.validate()
                except SiriusValidationError as e:
                    raise StateMachineTerminatedWithError(REQUEST_NOT_ACCEPTED, e.message)
                cred_infos, schemas, credential_defs, rev_states = await self._extract_credentials_info(
                    request.proof_request, self.__pool_name
                )

                if cred_infos.get('requested_attributes', None) or cred_infos.get('requested_predicates', None):
                    # Step-2: Build proof
                    proof = await sirius_sdk.AnonCreds.prover_create_proof(
                        proof_req=request.proof_request,
                        requested_credentials=cred_infos,
                        master_secret_name=master_secret_id,
                        schemas=schemas,
                        credential_defs=credential_defs,
                        rev_states=rev_states
                    )
                    # Step-3: Send proof and wait Ack to check success from Verifier side
                    presentation_msg = PresentationMessage(proof, version=request.version)
                    presentation_msg.please_ack = True
                    if request.please_ack:
                        presentation_msg.thread_id = request.ack_message_id

                    # Step-3: Wait ACK
                    await self.log(progress=50, message='Send presentation')

                    # Switch to await participant action
                    ack = await self.switch(presentation_msg)

                    if isinstance(ack, Ack):
                        await self.log(progress=100, message='Verify OK!')
                        return True
                    elif isinstance(ack, PresentProofProblemReport):
                        await self.log(progress=100, message='Verify ERROR!')
                        return False
                    else:
                        raise StateMachineTerminatedWithError(
                            RESPONSE_FOR_UNKNOWN_REQUEST, 'Unexpected response @type: %s' % str(ack.type)
                        )
                else:
                    raise StateMachineTerminatedWithError(
                        REQUEST_PROCESSING_ERROR, 'No proof correspondent to proof-request'
                    )
            except StateMachineTerminatedWithError as e:
                self._problem_report = PresentProofProblemReport(
                    e.problem_code, e.explain
                )
                await self.log(
                    progress=100, message=f'Terminated with error',
                    problem_code=e.problem_code, explain=e.explain
                )
                if e.notify:
                    await self.send(self._problem_report)
                return False

    async def _extract_credentials_info(self, proof_request, pool_name: str) -> (dict, dict, dict, dict):
        # Extract credentials from wallet that satisfy to request
        proof_response = await sirius_sdk.AnonCreds.prover_search_credentials_for_proof_req(
            proof_request, limit_referents=1
        )
        schemas = {}
        credential_defs = {}
        rev_states = {}
        opts = CacheOptions()
        requested_credentials = {
            'self_attested_attributes': {},
            'requested_attributes': {},
            'requested_predicates': {}
        }
        all_infos = []
        for referent_id, cred_infos in proof_response['requested_attributes'].items():
            cred_info = cred_infos[0]['cred_info']  # Get first
            info = {
                'cred_id': cred_info['referent'],
                'revealed': True
            }
            requested_credentials['requested_attributes'][referent_id] = info
            all_infos.append(cred_info)
        for referent_id, predicates in proof_response['requested_predicates'].items():
            pred_info = predicates[0]['cred_info']  # Get first
            info = {
                'cred_id': pred_info['referent']
            }
            requested_credentials['requested_predicates'][referent_id] = info
            all_infos.append(pred_info)
        for cred_info in all_infos:
            schema_id = cred_info['schema_id']
            cred_def_id = cred_info['cred_def_id']
            schema = await sirius_sdk.Cache.get_schema(
                pool_name=pool_name, submitter_did=self.__verifier.me.did,
                id_=schema_id, options=opts
            )
            cred_def = await sirius_sdk.Cache.get_cred_def(
                pool_name=pool_name, submitter_did=self.__verifier.me.did,
                id_=cred_def_id, options=opts
            )
            schemas[schema_id] = schema
            credential_defs[cred_def_id] = cred_def
        return requested_credentials, schemas, credential_defs, rev_states

    def _is_leader(self) -> bool:
        return False



import os
from setuptools import setup, find_packages

HERE = os.path.abspath(os.path.dirname(__file__))

# The text of the README file
with open(os.path.join(HERE, "README.rst")) as fid:
    README = fid.read()


setup(
    name="sirius-sdk",
    version="1.2.2",  # + os.getenv('BUILD_NUMBER'),
    description="SDK for developing Smart-Contracts in Self-sovereign-identity world.",
    long_description=README,
    long_description_content_type="text/x-rst",
    url="https://github.com/Sirius-social/sirius-sdk-python",
    author="Networks Synergy",
    author_email="support@socialsirius.com",
    project_urls={
        'Chat: Telegram': 'https://t.me/sirius_sdk',
        'CI: Travis': 'https://travis-ci.com/github/Sirius-social/sirius-sdk-python/builds',
        'GitHub: issues': 'https://github.com/Sirius-social/sirius-sdk-python/issues',
        'GitHub: repo': 'https://github.com/Sirius-social/sirius-sdk-python',
    },
    license="Apache License",
    maintainer=', '.join(('Pavel Minenkov <minikspb@gmail.com>',)),
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
    ],
    packages=find_packages(),
    python_requires='>=3.7',
    include_package_data=True,
    install_requires=[
        'aiohttp>=3.7.4',
        'base58>=2.0.0',
        'multipledispatch==0.6.0',
        'PyNaCl==1.3.0',
        'python-dateutil>=2.8.1',
        'pytime>=0.2.0',
        'semver>=2.10.1',
        'sortedcontainers~=2.3.0'
    ]
)



import uuid

import pytest

import sirius_sdk
from sirius_sdk.agent.wallet.abstract.non_secrets import RetrieveRecordOptions
from .helpers import ServerTestSuite


@pytest.mark.asyncio
async def test_record_value_ops(test_suite: ServerTestSuite):
    params = test_suite.get_agent_params('agent4')
    async with sirius_sdk.context(params['server_address'], params['credentials'], params['p2p']):
        value = 'my-value'
        my_id = 'my-id-' + uuid.uuid4().hex
        await sirius_sdk.NonSecrets.add_wallet_record('type', my_id, value)
        opts = RetrieveRecordOptions()
        opts.check_all()
        value_info = await sirius_sdk.NonSecrets.get_wallet_record('type', my_id, opts)
        assert value_info['id'] == my_id
        assert value_info['tags'] == {}
        assert value_info['value'] == value
        assert value_info['type'] == 'type'

        value_new = 'my-new-value'
        await sirius_sdk.NonSecrets.update_wallet_record_value('type', my_id, value_new)
        value_info = await sirius_sdk.NonSecrets.get_wallet_record('type', my_id, opts)
        assert value_info['value'] == value_new
        await sirius_sdk.NonSecrets.delete_wallet_record('type', my_id)


@pytest.mark.asyncio
async def test_record_tags_ops(test_suite: ServerTestSuite):
    params = test_suite.get_agent_params('agent4')
    async with sirius_sdk.context(params['server_address'], params['credentials'], params['p2p']):
        my_id = 'my-id-' + uuid.uuid4().hex

        value = 'my-value'
        tags = {
            'tag1': 'val1',
            '~tag2': 'val2'
        }
        await sirius_sdk.NonSecrets.add_wallet_record('type', my_id, value, tags)
        opts = RetrieveRecordOptions()
        opts.check_all()
        value_info = await sirius_sdk.NonSecrets.get_wallet_record('type', my_id, opts)
        assert value_info['id'] == my_id
        assert value_info['tags'] == tags
        assert value_info['value'] == value
        assert value_info['type'] == 'type'

        upd_tags = {
            'ext-tag': 'val3'
        }
        await sirius_sdk.NonSecrets.update_wallet_record_tags('type', my_id, upd_tags)
        value_info = await sirius_sdk.NonSecrets.get_wallet_record('type', my_id, opts)
        assert value_info['tags'] == upd_tags

        await sirius_sdk.NonSecrets.add_wallet_record_tags('type', my_id, tags)
        value_info = await sirius_sdk.NonSecrets.get_wallet_record('type', my_id, opts)
        assert value_info['tags'] == dict(**upd_tags, **tags)

        await sirius_sdk.NonSecrets.delete_wallet_record_tags('type', my_id, ['ext-tag'])
        value_info = await sirius_sdk.NonSecrets.get_wallet_record('type', my_id, opts)
        assert value_info['tags'] == tags


@pytest.mark.asyncio
async def test_maintain_tags_only_update_ops(test_suite: ServerTestSuite):
    params = test_suite.get_agent_params('agent4')
    async with sirius_sdk.context(params['server_address'], params['credentials'], params['p2p']):
        my_id = 'my-id-' + uuid.uuid4().hex
        value = 'my-value'
        await sirius_sdk.NonSecrets.add_wallet_record('type', my_id, value, )
        opts = RetrieveRecordOptions()
        opts.check_all()
        value_info = await sirius_sdk.NonSecrets.get_wallet_record('type', my_id, opts)
        assert value_info['id'] == my_id
        assert value_info['tags'] == {}
        assert value_info['value'] == value
        assert value_info['type'] == 'type'

        tags1 = {
            'tag1': 'val1',
            '~tag2': 'val2'
        }

        await sirius_sdk.NonSecrets.update_wallet_record_tags('type', my_id, tags1)
        value_info = await sirius_sdk.NonSecrets.get_wallet_record('type', my_id, opts)
        assert value_info['tags'] == tags1

        tags2 = {
            'tag3': 'val3',
        }
        await sirius_sdk.NonSecrets.update_wallet_record_tags('type', my_id, tags2)
        value_info = await sirius_sdk.NonSecrets.get_wallet_record('type', my_id, opts)
        assert value_info['tags'] == tags2


@pytest.mark.asyncio
async def test_wallet_search_sqlite(test_suite: ServerTestSuite):
    params = test_suite.get_agent_params('agent4')
    async with sirius_sdk.context(params['server_address'], params['credentials'], params['p2p']):
        my_id1 = 'my-id-' + uuid.uuid4().hex
        my_id2 = 'my-id-' + uuid.uuid4().hex
        type_ = 'type_' + uuid.uuid4().hex
        opts = RetrieveRecordOptions()
        opts.check_all()
        tags1 = {
            'tag1': 'val1',
            '~tag2': '5',
            'marker': 'A'
        }
        tags2 = {
            'tag3': 'val3',
            '~tag4': '6',
            'marker': 'B'
        }
        await sirius_sdk.NonSecrets.add_wallet_record(type_, my_id1, 'value1', tags1)
        await sirius_sdk.NonSecrets.add_wallet_record(type_, my_id2, 'value2', tags2)

        query = {
            "tag1": "val1"
        }
        records, total = await sirius_sdk.NonSecrets.wallet_search(type_, query, opts)
        assert total == 1
        assert 'value1' in str(records)

        query = {
            "$or": [{"tag1": "val1"}, {"~tag4": "6"}]
        }
        records, total = await sirius_sdk.NonSecrets.wallet_search(type_, query, opts)
        assert len(records) == 1
        assert total == 2

        records, total = await sirius_sdk.NonSecrets.wallet_search(type_, query, opts, limit=1000)
        assert len(records) == 2
        assert total == 2

        query = {
            "marker": {"$in": ["A", "C"]}
        }
        records, total = await sirius_sdk.NonSecrets.wallet_search(type_, query, opts)
        assert 1 == total



import json
from typing import Union, List, Dict

from sirius_sdk.agent.connections import AgentRPC
from sirius_sdk.agent.microledgers.abstract import AbstractMicroledgerList, Transaction, AbstractMicroledger, \
    LedgerMeta, MerkleInfo, AuditProof, AbstractBatchedAPI
from sirius_sdk.errors.exceptions import SiriusContextError
from sirius_sdk.agent.microledgers.expiringdict import ExpiringDict


class BatchedAPI(AbstractBatchedAPI):

    def __init__(self, api: AgentRPC, external: Dict[str, AbstractMicroledger] = None):
        self.__api = api
        self.__names = []
        self.__external = external

    async def open(self, ledgers: Union[List[str], List[AbstractMicroledger]]) -> List[AbstractMicroledger]:
        names_to_open = []
        for ledger in ledgers:
            if isinstance(ledger, AbstractMicroledger):
                names_to_open.append(ledger.name)
            elif isinstance(ledger, str):
                names_to_open.append(ledger)
            else:
                raise RuntimeError(f'Unexpected ledgers item type: {str(type(ledger))}')

        names_to_open = list(set(names_to_open))  # remove duplicates
        await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers-batched/1.0/open',
            params={
                'names': names_to_open
            }
        )
        self.__names = names_to_open
        states = await self.states()
        return states

    async def close(self):
        await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers-batched/1.0/close'
        )

    async def states(self) -> List[AbstractMicroledger]:
        states = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers-batched/1.0/states'
        )
        resp = self.__return_ledgers(states)
        return resp

    async def append(
            self, transactions: Union[List[Transaction], List[dict]], txn_time: Union[str, int] = None
    ) -> List[AbstractMicroledger]:
        states = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers-batched/1.0/append_txns',
            params={
                'txns': transactions,
                'txn_time': txn_time
            }
        )
        resp = self.__return_ledgers(states)
        return resp

    async def commit(self) -> List[AbstractMicroledger]:
        states = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers-batched/1.0/commit_txns'
        )
        resp = self.__return_ledgers(states)
        return resp

    async def reset_uncommitted(self) -> List[AbstractMicroledger]:
        states = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers-batched/1.0/reset_uncommitted'
        )
        resp = self.__return_ledgers(states)
        return resp

    def __return_ledgers(self, states: dict) -> List[AbstractMicroledger]:
        resp = []  # keep ledgers ordering
        for name in self.__names:
            state = states[name]
            ledger = Microledger(name, self.__api, state)
            if self.__external is not None:
                if name in self.__external:
                    ledger.assign_to(self.__external[name])
                else:
                    self.__external[name] = ledger
            resp.append(ledger)
        return resp


class MicroledgerList(AbstractMicroledgerList):

    TTL = 60*60  # 1 hour

    def __init__(self, api: AgentRPC):
        self.__api = api
        self.instances: Dict[str, AbstractMicroledger] = ExpiringDict(ttl=self.TTL)
        self.__batched_api = BatchedAPI(api, self.instances)

    async def batched(self) -> AbstractBatchedAPI:
        return self.__batched_api

    async def create(self, name: str, genesis: Union[List[Transaction], List[dict]]) -> (AbstractMicroledger, List[Transaction]):
        genesis_txns = []
        for txn in genesis:
            if isinstance(txn, Transaction):
                genesis_txns.append(txn)
            elif isinstance(txn, dict):
                genesis_txns.append(Transaction.create(txn))
            else:
                raise RuntimeError('Unexpected transaction type')
        instance = Microledger(name, self.__api)
        txns = await instance.init(genesis_txns)
        self.instances[name] = instance
        return instance, txns

    async def ledger(self, name: str) -> AbstractMicroledger:
        if name not in self.instances:
            await self.__check_is_exists(name)
            instance = Microledger(name, self.__api)
            self.instances[name] = instance
        return self.instances[name]

    async def reset(self, name: str):
        await self.__check_is_exists(name)
        await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/reset',
            params={
                'name': name
            }
        )
        if name in self.instances.keys():
            del self.instances[name]

    async def is_exists(self, name: str):
        is_exists = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/is_exists',
            params={
                'name': name
            }
        )
        return is_exists

    async def leaf_hash(self, txn: Union[Transaction, bytes]) -> bytes:
        if isinstance(txn, Transaction):
            data = json.dumps(txn, sort_keys=True, ensure_ascii=False, separators=(',', ':')).encode()
        elif isinstance(txn, bytes):
            data = txn
        else:
            raise RuntimeError('Unexpected transaction type')
        leaf_hash = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/leaf_hash',
            params={
                'data': data
            }
        )
        return leaf_hash

    async def list(self) -> List[LedgerMeta]:
        collection = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/list',
            params={
                'name': '*'
            }
        )
        return [LedgerMeta(**item) for item in collection]

    async def __check_is_exists(self, name: str):
        if name not in self.instances.keys():
            is_exists = await self.is_exists(name)
            if not is_exists:
                raise SiriusContextError('MicroLedger with name "" does not exists' % name)


class Microledger(AbstractMicroledger):

    def __init__(self, name: str, api: AgentRPC, state: dict = None):
        self.__name = name
        self.__api = api
        self.__state = state

    def assign_to(self, other: AbstractMicroledger):
        if isinstance(other, Microledger):
            other.__state = self.__state

    @property
    def name(self) -> str:
        return self.__name

    @property
    def size(self) -> int:
        self.__check_state_is_exists()
        return self.__state['size']

    @property
    def uncommitted_size(self) -> int:
        self.__check_state_is_exists()
        return self.__state['uncommitted_size']

    @property
    def root_hash(self) -> str:
        self.__check_state_is_exists()
        return self.__state['root_hash']

    @property
    def uncommitted_root_hash(self) -> str:
        self.__check_state_is_exists()
        return self.__state['uncommitted_root_hash']

    @property
    def seq_no(self) -> int:
        self.__check_state_is_exists()
        return self.__state['seqNo']

    async def reload(self):
        state = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/state',
            params={
                'name': self.name
            }
        )
        self.__state = state

    async def rename(self, new_name: str):
        await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/rename',
            params={
                'name': self.name,
                'new_name': new_name
            }
        )
        self.__name = new_name

    async def init(self, genesis: List[Transaction]) -> List[Transaction]:
        self.__state, txns = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/initialize',
            params={
                'name': self.name,
                'genesis_txns': genesis
            }
        )
        txns = [Transaction.from_value(txn) for txn in txns]
        return txns

    async def append(
            self, transactions: Union[List[Transaction], List[dict]], txn_time: Union[str, int] = None
    ) -> (int, int, List[Transaction]):
        transactions_to_append = []
        for txn in transactions:
            if isinstance(txn, Transaction):
                transactions_to_append.append(txn)
            elif isinstance(txn, dict):
                transactions_to_append.append(Transaction.create(txn))
            else:
                raise RuntimeError('Unexpected transaction type')
        transactions_with_meta = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/append_txns_metadata',
            params={
                'name': self.name,
                'txns': transactions_to_append,
                'txn_time': txn_time
            }
        )
        self.__state, start, end, appended_txns = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/append_txns',
            params={
                'name': self.name,
                'txns': transactions_with_meta,
            }
        )
        return start, end, Transaction.from_value(appended_txns)

    async def commit(self, count: int) -> (int, int, List[Transaction]):
        self.__state, start, end, committed_txns = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/commit_txns',
            params={
                'name': self.name,
                'count': count,
            }
        )
        return start, end, Transaction.from_value(committed_txns)

    async def discard(self, count: int):
        self.__state = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/discard_txns',
            params={
                'name': self.name,
                'count': count,
            }
        )

    async def merkle_info(self, seq_no: int) -> MerkleInfo:
        merkle_info = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/merkle_info',
            params={
                'name': self.name,
                'seqNo': seq_no,
            }
        )
        return MerkleInfo(
            root_hash=merkle_info['rootHash'],
            audit_path=merkle_info['auditPath'],
        )

    async def audit_proof(self, seq_no: int) -> AuditProof:
        proof = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/audit_proof',
            params={
                'name': self.name,
                'seqNo': seq_no,
            }
        )
        return AuditProof(
            root_hash=proof['rootHash'],
            audit_path=proof['auditPath'],
            ledger_size=proof['ledgerSize']
        )

    async def reset_uncommitted(self):
        self.__state = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/reset_uncommitted',
            params={
                'name': self.name,
            }
        )

    async def get_transaction(self, seq_no: int) -> Transaction:
        txn = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/get_by_seq_no',
            params={
                'name': self.name,
                'seqNo': seq_no
            }
        )
        txn = Transaction.from_value(txn)
        assert isinstance(txn, Transaction)
        return txn

    async def get_uncommitted_transaction(self, seq_no: int) -> Transaction:
        txn = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/get_by_seq_no_uncommitted',
            params={
                'name': self.name,
                'seqNo': seq_no
            }
        )
        txn = Transaction.from_value(txn)
        assert isinstance(txn, Transaction)
        return txn

    async def get_last_transaction(self) -> Transaction:
        txn = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/get_last_txn',
            params={
                'name': self.name
            }
        )
        txn = Transaction.from_value(txn)
        assert isinstance(txn, Transaction)
        return txn

    async def get_last_committed_transaction(self) -> Transaction:
        txn = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/get_last_committed_txn',
            params={
                'name': self.name
            }
        )
        txn = Transaction.from_value(txn)
        assert isinstance(txn, Transaction)
        return txn

    async def get_all_transactions(self) -> List[Transaction]:
        txns = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/get_all_txns',
            params={
                'name': self.name
            }
        )
        txns = [t[1] for t in txns]
        txns = Transaction.from_value(txns)
        assert isinstance(txns, list)
        return txns

    async def get_uncommitted_transactions(self) -> List[Transaction]:
        txns = await self.__api.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/microledgers/1.0/get_uncommitted_txns',
            params={
                'name': self.name
            }
        )
        txns = Transaction.from_value(txns)
        assert isinstance(txns, list)
        return txns

    def __check_state_is_exists(self):
        if self.__state is None:
            raise SiriusContextError('Load state of Microledger at First!')


import os
import json
import asyncio
from typing import List, Any
from urllib.parse import urljoin, urlparse

import aiohttp
import pytest

from sirius_sdk import Agent, Pairwise
from sirius_sdk.base import ReadOnlyChannel, WriteOnlyChannel
from sirius_sdk.errors.exceptions import SiriusTimeoutIO
from sirius_sdk.encryption import P2PConnection
from sirius_sdk.agent.aries_rfc.feature_0036_issue_credential import ProposedAttrib as IssuingProposedAttrib, \
    AttribTranslation as IssuingAttribTranslation


class IndyAgent:

    WALLET = 'test'
    PASS_PHRASE = 'pass'
    DEFAULT_LABEL = 'BackCompatibility'
    SETUP_TIMEOUT = 60

    def __init__(self):
        self.__address = pytest.old_agent_address
        self.__auth_username = pytest.old_agent_root['username']
        self.__auth_password = pytest.old_agent_root['password']
        self.__endpoint = None
        self.__wallet_exists = False
        self.__endpoint = None
        self.__default_invitation = None

    @property
    def endpoint(self) -> dict:
        return self.__endpoint

    @property
    def default_invitation(self) -> dict:
        return self.__default_invitation

    async def invite(self, invitation_url: str, for_did: str=None, ttl: int=None):
        url = '/agent/admin/wallets/%s/endpoints/%s/invite/' % (self.WALLET, self.endpoint['uid'])
        params = {'url': invitation_url, 'pass_phrase': self.PASS_PHRASE}
        if for_did:
            params['my_did'] = for_did
        if ttl:
            params['ttl'] = ttl
        ok, resp = await self.__http_post(
            path=url,
            json_=params
        )
        assert ok

    async def load_invitations(self):
        url = '/agent/admin/wallets/%s/endpoints/%s/invitations/' % (self.WALLET, self.__endpoint['uid'])
        ok, collection = await self.__http_get(url)
        assert ok is True
        return collection

    async def create_invitation(self, label: str, seed: str=None):
        url = '/agent/admin/wallets/%s/endpoints/%s/invitations/' % (self.WALLET, self.__endpoint['uid'])
        params = {'label': label, 'pass_phrase': self.PASS_PHRASE}
        if seed:
            params['seed'] = seed
        ok, invitation = await self.__http_post(url, params)
        assert ok is True
        return invitation

    async def create_and_store_my_did(self, seed: str = None) -> (str, str):
        url = '/agent/admin/wallets/%s/did/create_and_store_my_did/' % self.WALLET
        params = {'pass_phrase': self.PASS_PHRASE}
        if seed:
            params['seed'] = seed
        ok, resp = await self.__http_post(url, params)
        assert ok is True
        return resp['did'], resp['verkey']

    async def create_pairwise_statically(self, pw: Pairwise):
        url = '/agent/admin/wallets/%s/pairwise/create_pairwise_statically/' % self.WALLET
        metadata = {
            'label': pw.their.label,
            'their_vk': pw.their.verkey,
            'my_vk': pw.me.verkey,
            'their_endpoint': pw.their.endpoint
        }
        params = {'pass_phrase': self.PASS_PHRASE}
        params.update({
            'my_did': pw.me.did,
            'their_did': pw.their.did,
            'their_verkey': pw.their.verkey,
            'metadata': metadata
        })
        ok, resp = await self.__http_post(url, params)
        assert ok is True

    async def register_schema(self, issuer_did: str, name: str, version: str, attributes: List[str]) -> (str, dict):
        url = '/agent/admin/wallets/%s/did/%s/ledger/register_schema/' % (self.WALLET, issuer_did)
        params = {
            'pass_phrase': self.PASS_PHRASE,
            'name': name,
            'version': version,
            'attributes': attributes
        }
        ok, resp = await self.__http_post(url, params)
        assert ok is True
        return resp['schema_id'], resp['schema']

    async def register_cred_def(
            self, submitter_did: str, schema_id : str, tag: str, support_revocation: bool = False
    ):
        url = '/agent/admin/wallets/%s/did/%s/cred_def/create_and_send/' % (self.WALLET, submitter_did)
        params = {
            'pass_phrase': self.PASS_PHRASE,
            'schema_id': schema_id,
            'tag': tag,
            'support_revocation': support_revocation
        }
        ok, resp = await self.__http_post(url, params)
        assert ok is True
        return resp['id'], resp['cred_def']

    async def issue_credential(
            self, cred_def_id: str, cred_def: dict, values: dict, their_did: str,
            comment: str = None, locale: str = None, issuer_schema: dict = None,
            preview: List[IssuingProposedAttrib] = None, translation: List[IssuingAttribTranslation] = None,
            rev_reg_id: str = None, cred_id: str = None, ttl: int = 60
    ) -> Any:
        url = '/agent/admin/wallets/%s/messaging/issue_credential/' % self.WALLET
        params = {
            'pass_phrase': self.PASS_PHRASE,
            'cred_def_id': cred_def_id,
            'cred_def': cred_def,
            'values': values,
            'their_did': their_did
        }
        if comment:
            params['comment'] = comment
        if locale:
            params['locale'] = locale
        if issuer_schema:
            params['issuer_schema'] = issuer_schema
        if preview:
            params['preview'] = preview
        if translation:
            params['translation'] = translation
        if rev_reg_id:
            params['rev_reg_id'] = rev_reg_id
        if cred_id:
            params['cred_id'] = cred_id
        if ttl:
            params['ttl'] = ttl
        params['collect_log'] = True
        ok, resp = await self.__http_post(url, params)
        assert ok is True
        return resp

    async def ensure_is_alive(self):
        inc_timeout = 10
        for n in range(1, self.SETUP_TIMEOUT, inc_timeout):
            ok, wallets = await self.__http_get('/agent/admin/wallets/')
            if ok:
                break
            progress = float(n / self.SETUP_TIMEOUT) * 100
            print('Indy-Agent setup Progress: %.1f %%' % progress)
            await asyncio.sleep(inc_timeout)
        if not self.__wallet_exists:
            ok, wallets = await self.__http_post(
                '/agent/admin/wallets/ensure_exists/',
                {'uid': self.WALLET, 'pass_phrase': self.PASS_PHRASE}
            )
            assert ok is True
            self.__wallet_exists = True
        ok, resp = await self.__http_post(
            '/agent/admin/wallets/%s/open/' % self.WALLET,
            {'pass_phrase': self.PASS_PHRASE}
        )
        assert ok
        if not self.__endpoint:
            url = '/agent/admin/wallets/%s/endpoints/' % self.WALLET
            ok, resp = await self.__http_get(url)
            assert ok is True
            if resp['results']:
                self.__endpoint = resp['results'][0]
            else:
                ok, endpoint = ok, wallets = await self.__http_post(url, {'host': self.__address})
                assert ok is True
                self.__endpoint = endpoint
        if not self.__default_invitation:
            url = '/agent/admin/wallets/%s/endpoints/%s/invitations/' % (self.WALLET, self.__endpoint['uid'])
            ok, resp = await self.__http_get(url)
            assert ok is True
            collection = [item for item in resp if item['seed'] == 'default']
            if collection:
                self.__default_invitation = collection[0]
            else:
                ok, invitaion = ok, wallets = await self.__http_post(
                    url,
                    {'label': self.DEFAULT_LABEL, 'pass_phrase': self.PASS_PHRASE, 'seed': 'default'}
                )
                assert ok is True
                self.__default_invitation = invitaion

    async def __http_get(self, path: str):
        url = urljoin(self.__address, path)
        auth = aiohttp.BasicAuth(self.__auth_username, self.__auth_password, 'utf-8')
        netloc = urlparse(self.__address).netloc
        host = netloc.split(':')[0]
        async with aiohttp.ClientSession(auth=auth) as session:
            headers = {
                'content-type': 'application/json',
                'host': host
            }
            try:
                async with session.get(url, headers=headers) as resp:
                    if resp.status in [200]:
                        content = await resp.json()
                        return True, content
                    else:
                        err_message = await resp.text()
                        return False, err_message
            except aiohttp.ClientError:
                return False, None

    async def __http_post(self, path: str, json_: dict = None):
        url = urljoin(self.__address, path)
        auth = aiohttp.BasicAuth(self.__auth_username, self.__auth_password, 'utf-8')
        netloc = urlparse(self.__address).netloc
        host = netloc.split(':')[0]
        async with aiohttp.ClientSession(auth=auth) as session:
            headers = {
                'content-type': 'application/json',
                'host': host
            }
            try:
                body = json.dumps(json_).encode() if json_ else None
                async with session.post(url, headers=headers, data=body) as resp:
                    if resp.status in [200, 201]:
                        try:
                            content = await resp.json()
                        except Exception as e:
                            content = None
                        return True, content
                    else:
                        err_message = await resp.text()
                        return False, err_message
            except aiohttp.ClientError:
                return False, None


class ServerTestSuite:

    SETUP_TIMEOUT = 60

    def __init__(self):
        self.__server_address = pytest.test_suite_baseurl
        self.__url = urljoin(self.__server_address, '/test_suite')
        self.__metadata = None
        test_suite_path = os.getenv('TEST_SUITE', None)
        if test_suite_path is None:
            self.__test_suite_exists_locally = False
        else:
            self.__test_suite_exists_locally = os.path.isfile(test_suite_path) and 'localhost' in self.__server_address

    @property
    def metadata(self):
        return self.__metadata
    
    def get_agent_params(self, name: str):
        if not self.__metadata:
            raise RuntimeError('TestSuite is not running...')
        agent = self.__metadata.get(name, None)
        if not agent:
            raise RuntimeError('TestSuite does not have agent with name "%s"' % name)
        p2p = agent['p2p']
        return {
            'server_address': self.__server_address,
            'credentials': agent['credentials'].encode('ascii'),
            'p2p': P2PConnection(
                my_keys=(
                    p2p['smart_contract']['verkey'],
                    p2p['smart_contract']['secret_key']
                ),
                their_verkey=p2p['agent']['verkey']
            ),
            'entities': agent['entities']
        }

    async def ensure_is_alive(self):
        ok, meta = await self.__http_get(self.__url)
        if ok:
            self.__metadata = meta
        else:
            if self.__test_suite_exists_locally:
                await self.__run_suite_locally()
            inc_timeout = 10
            print('\n\nStarting test suite locally...\n\n')

            for n in range(1, self.SETUP_TIMEOUT, inc_timeout):
                progress = float(n / self.SETUP_TIMEOUT)*100
                print('TestSuite setup progress: %.1f %%' % progress)
                await asyncio.sleep(inc_timeout)
                ok, meta = await self.__http_get(self.__url)
                if ok:
                    self.__metadata = meta
                    print('Server test suite was detected')
                    return
            print('Timeout for waiting TestSuite is alive expired!')
            raise RuntimeError('Expect server with running TestSuite. See conftest.py: pytest_configure')

    @staticmethod
    async def __run_suite_locally():
        os.popen('python /app/configure.py --asgi_port=$ASGI_PORT --wsgi_port=$WSGI_PORT --nginx_port=$PORT')
        await asyncio.sleep(1)
        os.popen('python /app/manage.py test_suite > /tmp/test_suite.log 2> /tmp/test_suite.err')
        os.popen('supervisord -c /etc/supervisord.conf & sudo nginx -g "daemon off;"')
        await asyncio.sleep(5)

    @staticmethod
    async def __http_get(url: str):
        async with aiohttp.ClientSession() as session:
            headers = {
                'content-type': 'application/json'
            }
            try:
                async with session.get(url, headers=headers) as resp:
                    if resp.status in [200]:
                        content = await resp.json()
                        return True, content
                    else:
                        err_message = await resp.text()
                        return False, err_message
            except aiohttp.ClientError:
                return False, None


class InMemoryChannel(ReadOnlyChannel, WriteOnlyChannel):

    def __init__(self):
        self.queue = asyncio.Queue()

    async def read(self, timeout: int = None) -> bytes:

        ret = None

        async def internal_reading():
            nonlocal ret
            ret = await self.queue.get()

        done, pending = await asyncio.wait([internal_reading()], timeout=timeout)

        for coro in pending:
            coro.cancel()
        if isinstance(ret, bytes):
            return ret
        else:
            raise SiriusTimeoutIO()

    async def write(self, data: bytes) -> bool:
        await self.queue.put(data)
        return True


async def run_coroutines(*args, timeout: int = 15):
    results = []
    items = [i for i in args]
    done, pending = await asyncio.wait(items, timeout=timeout, return_when=asyncio.FIRST_EXCEPTION)
    for f in done:
        if f.exception():
            raise f.exception()
        results.append(f.result())
    for f in pending:
        f.cancel()
    return results



from abc import ABC
from typing import List, Optional
from datetime import datetime, timedelta

from sirius_sdk.errors.exceptions import *
from sirius_sdk.messaging import *
from sirius_sdk.messaging.fields import DIDField
from sirius_sdk.agent.wallet.wallets import DynamicWallet
from sirius_sdk.agent.connections import AgentRPC, RoutingBatch
from sirius_sdk.agent.pairwise import TheirEndpoint, Pairwise
from sirius_sdk.agent.pairwise import AbstractPairwiseList, WalletPairwiseList


class AbstractCoProtocolTransport(ABC):
    """Abstraction application-level protocols in the context of interactions among agent-like things.

        Sirius SDK protocol is high-level abstraction over Sirius transport architecture.
        Approach advantages:
          - developer build smart-contract logic in block-style that is easy to maintain and control
          - human-friendly source code of state machines in procedural style
          - program that is running in separate coroutine: lightweight abstraction to start/kill/state-detection work thread
        See details:
          - https://github.com/hyperledger/aries-rfcs/tree/master/concepts/0003-protocols
        """

    THREAD_DECORATOR = '~thread'
    PLEASE_ACK_DECORATOR = '~please_ack'

    SEC_PER_DAY = 86400
    SEC_PER_HOURS = 3600
    SEC_PER_MIN = 60

    def __init__(self, rpc: AgentRPC):
        """
        :param rpc: RPC (independent connection)
        """
        self.__time_to_live = None
        self._rpc = rpc
        self._check_protocols = True
        self._check_verkeys = False
        self.__default_timeout = rpc.timeout
        self.__wallet = DynamicWallet(self._rpc)
        self.__pairwise_list = WalletPairwiseList(api=(self.__wallet.pairwise, self.__wallet.did))
        self.__die_timestamp = None
        self.__their_vk = None
        self.__endpoint = None
        self.__my_vk = None
        self.__routing_keys = None
        self.__is_setup = False
        self.__protocols = []
        self.__please_ack_ids = []
        self.__is_started = False

    @property
    def protocols(self) -> List[str]:
        return self.__protocols

    @property
    def time_to_live(self) -> int:
        return self.__time_to_live

    @property
    def is_started(self) -> bool:
        return self.__is_started

    def _setup(self, their_verkey: str, endpoint: str, my_verkey: str=None, routing_keys: List[str]=None):
        """Should be called in Descendant"""
        self.__their_vk = their_verkey
        self.__my_vk = my_verkey
        self.__endpoint = endpoint
        self.__routing_keys = routing_keys or []
        self.__is_setup = True

    @property
    def wallet(self) -> DynamicWallet:
        return self.__wallet

    @property
    def pairwise_list(self) -> AbstractPairwiseList:
        return self.__pairwise_list

    @property
    def is_alive(self) -> bool:
        if self.__die_timestamp:
            return datetime.now() < self.__die_timestamp
        else:
            return False

    async def start(self, protocols: List[str], time_to_live: int=None):
        self.__protocols = protocols
        self.__time_to_live = time_to_live
        if self.__time_to_live:
            self.__die_timestamp = datetime.now() + timedelta(seconds=self.__time_to_live)
        else:
            self.__die_timestamp = None
        self.__is_started = True

    async def stop(self):
        self.__die_timestamp = None
        self.__is_started = False
        await self.__cleanup_context()

    async def switch(self, message: Message) -> (bool, Message):
        """Send Message to other-side of protocol and wait for response

        :param message: Protocol request
        :return: (success, Response)
        """
        if not self.__is_setup:
            raise SiriusPendingOperation('You must Setup protocol instance at first')
        try:
            self._rpc.timeout = self.__get_io_timeout()
            await self.__setup_context(message)
            try:
                event = await self._rpc.send_message(
                    message=message,
                    their_vk=self.__their_vk,
                    endpoint=self.__endpoint,
                    my_vk=self.__my_vk,
                    routing_keys=self.__routing_keys,
                    coprotocol=True
                )
            finally:
                await self.__cleanup_context(message)
            if self._check_verkeys:
                recipient_verkey = event.get('recipient_verkey', None)
                sender_verkey = event.get('sender_verkey')
                if recipient_verkey != self.__my_vk:
                    raise SiriusInvalidPayloadStructure(f'Unexpected recipient_verkey: {recipient_verkey}')
                if sender_verkey != self.__their_vk:
                    raise SiriusInvalidPayloadStructure(f'Unexpected sender_verkey: {sender_verkey}')
            payload = Message(event.get('message', {}))
            if payload:
                ok, message = restore_message_instance(payload)
                if not ok:
                    message = Message(payload)
                if self._check_protocols:
                    if Type.from_str(message.type).protocol not in self.protocols:
                        raise SiriusInvalidMessage('@type has unexpected protocol "%s"' % message.type.protocol)
                return True, message
            else:
                return False, None
        except SiriusTimeoutIO:
            return False, None

    async def get_one(self) -> (Optional[Message], str, Optional[str]):
        timeout = self.__get_io_timeout()
        if (timeout is not None) and (timeout <= 0):
            raise SiriusTimeoutIO
        self._rpc.timeout = timeout
        event = await self._rpc.read_protocol_message()
        if 'message' in event:
            ok, message = restore_message_instance(event['message'])
            if not ok:
                message = Message(event['message'])
        else:
            message = None
        sender_verkey = event.get('sender_verkey', None)
        recipient_verkey = event.get('recipient_verkey', None)
        return message, sender_verkey, recipient_verkey

    async def send(self, message: Message):
        """Send message and don't wait answer

        :param message:
        :return:
        """
        if not self.__is_setup:
            raise SiriusPendingOperation('You must Setup protocol instance at first')
        self._rpc.timeout = self.__get_io_timeout()
        await self.__setup_context(message)
        await self._rpc.send_message(
            message=message,
            their_vk=self.__their_vk,
            endpoint=self.__endpoint,
            my_vk=self.__my_vk,
            routing_keys=self.__routing_keys,
            coprotocol=False,
            ignore_errors=True
        )

    async def send_many(self, message: Message, to: List[Pairwise]) -> List[Any]:
        batches = [
            RoutingBatch(p.their.verkey, p.their.endpoint, p.me.verkey, p.their.routing_keys)
            for p in to
        ]
        if not self.__is_setup:
            raise SiriusPendingOperation('You must Setup protocol instance at first')
        self._rpc.timeout = self.__get_io_timeout()
        await self.__setup_context(message)
        results = await self._rpc.send_message_batched(
            message, batches
        )
        return results

    async def __setup_context(self, message: Message):
        if self.PLEASE_ACK_DECORATOR in message:
            ack_message_id = message.get(self.PLEASE_ACK_DECORATOR, {}).get('message_id', None) or message.id
            ttl = self.__get_io_timeout() or 3600
            await self._rpc.start_protocol_with_threads(
                threads=[ack_message_id], ttl=ttl
            )
            self.__please_ack_ids.append(ack_message_id)

    async def __cleanup_context(self, message: Message=None):
        if message:
            if self.PLEASE_ACK_DECORATOR in message:
                ack_message_id = message.get(self.PLEASE_ACK_DECORATOR, {}).get('message_id', None) or message.id
                await self._rpc.stop_protocol_with_threads(
                    threads=[ack_message_id], off_response=True
                )
                self.__please_ack_ids = [i for i in self.__please_ack_ids if i != ack_message_id]
        else:
            await self._rpc.stop_protocol_with_threads(
                threads=self.__please_ack_ids, off_response=True
            )
            self.__please_ack_ids.clear()

    def __get_io_timeout(self):
        if self.__die_timestamp:
            now = datetime.now()
            if now < self.__die_timestamp:
                delta = self.__die_timestamp - now
                return delta.days * self.SEC_PER_DAY + delta.seconds
            else:
                return 0
        else:
            return None


class TheirEndpointCoProtocolTransport(AbstractCoProtocolTransport):

    def __init__(
            self, my_verkey: str, endpoint: TheirEndpoint, rpc: AgentRPC
    ):
        super().__init__(rpc)
        self.__endpoint = endpoint
        self.__my_verkey = my_verkey
        self._setup(
            their_verkey=endpoint.verkey,
            endpoint=endpoint.endpoint,
            my_verkey=my_verkey,
            routing_keys=endpoint.routing_keys
        )

    async def start(self, protocols: List[str], time_to_live: int = None):
        await super().start(protocols, time_to_live)
        await self._rpc.start_protocol_for_p2p(
            sender_verkey=self.__my_verkey,
            recipient_verkey=self.__endpoint.verkey,
            protocols=self.protocols,
            ttl=time_to_live
        )

    async def stop(self):
        await super().stop()
        await self._rpc.stop_protocol_for_p2p(
            sender_verkey=self.__my_verkey,
            recipient_verkey=self.__endpoint.verkey,
            protocols=self.protocols,
            off_response=True
        )


class PairwiseCoProtocolTransport(AbstractCoProtocolTransport):

    def __init__(
            self, pairwise: Pairwise, rpc: AgentRPC
    ):
        super().__init__(rpc)
        self.__pairwise = pairwise
        self._setup(
            their_verkey=pairwise.their.verkey,
            endpoint=pairwise.their.endpoint,
            my_verkey=pairwise.me.verkey,
            routing_keys=pairwise.their.routing_keys
        )

    async def start(self, protocols: List[str], time_to_live: int=None):
        await super().start(protocols, time_to_live)
        await self._rpc.start_protocol_for_p2p(
            sender_verkey=self.__pairwise.me.verkey,
            recipient_verkey=self.__pairwise.their.verkey,
            protocols=self.protocols,
            ttl=time_to_live
        )

    async def stop(self):
        await super().stop()
        await self._rpc.stop_protocol_for_p2p(
            sender_verkey=self.__pairwise.me.verkey,
            recipient_verkey=self.__pairwise.their.verkey,
            protocols=self.protocols,
            off_response=True
        )


class ThreadBasedCoProtocolTransport(AbstractCoProtocolTransport):
    """CoProtocol based on ~thread decorator

    See details:
      - https://github.com/hyperledger/aries-rfcs/tree/master/concepts/0008-message-id-and-threading
    """

    def __init__(
            self, thid: str, pairwise: Optional[Pairwise], rpc: AgentRPC, pthid: str=None
    ):
        super().__init__(rpc)
        self.__thid = thid
        self.__pthid = pthid
        self.__sender_order = 0
        self.__received_orders = {}
        self.pairwise = pairwise

    @property
    def pairwise(self) -> Pairwise:
        return self.__pairwise

    @pairwise.setter
    def pairwise(self, value: Optional[Pairwise]):
        self.__pairwise = value
        if value:
            self.__their = value.their
            self._setup(
                their_verkey=value.their.verkey,
                endpoint=value.their.endpoint,
                my_verkey=value.me.verkey,
                routing_keys=value.their.routing_keys
            )
        else:
            self.__their = None
            self._setup('', '', None, None)

    async def start(self, protocols: List[str] = None, time_to_live: int = None):
        if protocols is None:
            self._check_protocols = False
        await super().start(protocols, time_to_live)
        await self._rpc.start_protocol_with_threading(self.__thid, time_to_live)

    async def stop(self):
        await super().stop()
        await self._rpc.stop_protocol_with_threading(self.__thid, True)

    async def switch(self, message: Message) -> (bool, Message):
        self.__prepare_message(message)
        ok, response = await super().switch(message)
        if ok:
            respond_sender_order = response.get('~thread', {}).get('sender_order', None)
            if respond_sender_order is not None and self.__their is not None:
                recipient = self.__their.did
                err = DIDField().validate(recipient)
                if err is None:
                    order = self.__received_orders.get(recipient, 0)
                    self.__received_orders[recipient] = max(order, respond_sender_order)
        return ok, response

    async def send(self, message: Message):
        self.__prepare_message(message)
        await super().send(message)

    async def send_many(self, message: Message, to: List[Pairwise]) -> List[Any]:
        self.__prepare_message(message)
        ret = await super().send_many(message, to)
        return ret

    def __prepare_message(self, message: Message):
        if self.THREAD_DECORATOR not in message:  # Don't rewrite existing ~thread decorator
            thread_decorator = {
                'thid': self.__thid,
                'sender_order': self.__sender_order
            }
            if self.__pthid:
                thread_decorator['pthid'] = self.__pthid
            if self.__received_orders:
                thread_decorator['received_orders'] = self.__received_orders
            self.__sender_order += 1
            message[self.THREAD_DECORATOR] = thread_decorator



import hashlib
import datetime
from typing import Optional, List

from sirius_sdk.agent.wallet.abstract.crypto import AbstractCrypto
from sirius_sdk.agent.aries_rfc.utils import sign
from sirius_sdk.agent.aries_rfc.mixins import ThreadMixin
from sirius_sdk.agent.aries_rfc.base import AriesProtocolMessage, RegisterMessage


class Question(ThreadMixin, AriesProtocolMessage, metaclass=RegisterMessage):
    """Implementation of Question

    https://github.com/hyperledger/aries-rfcs/tree/master/features/0113-question-answer
    """

    PROTOCOL = 'questionanswer'
    NAME = 'question'

    def __init__(
            self, valid_responses: List[str] = None, question_text: Optional[str] = None,
            question_detail: Optional[str] = None, nonce: Optional[str] = None,
            signature_required: Optional[bool] = None, locale: Optional[str] = None,
            expires_time: Optional[str] = None,
            *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        if locale is not None:
            self['~l10n'] = {"locale": locale}
        if valid_responses is not None:
            self['valid_responses'] = [{'text': s} for s in valid_responses]
        if question_text is not None:
            self['question_text'] = question_text
        if question_detail is not None:
            self['question_detail'] = question_detail
        if nonce is not None:
            self['nonce'] = nonce
        if signature_required is not None:
            self['signature_required'] = signature_required
        if expires_time:
            timing = self.get('~timing', {})
            timing['expires_time'] = expires_time
            self['~timing'] = timing

    @property
    def locale(self) -> Optional[str]:
        return self.get('~l10n', {}).get('locale', None)

    @property
    def valid_responses(self) -> List[str]:
        return list(self.get('valid_responses', {}).values())

    @property
    def question_text(self) -> Optional[str]:
        return self.get('question_text', None)

    @property
    def question_detail(self) -> Optional[str]:
        return self.get('question_detail', None)

    @property
    def nonce(self) -> Optional[str]:
        return self.get('nonce', None)

    @property
    def signature_required(self) -> Optional[bool]:
        return self.get('signature_required', None)

    @property
    def expires_time(self) -> Optional[str]:
        return self.get('~timing', {}).get('expires_time', None)

    def set_ttl(self, seconds: int):
        expire_at = datetime.datetime.utcnow() + datetime.timedelta(seconds=seconds)
        timing = self.get('~timing', {})
        timing['expires_time'] = expire_at.replace(tzinfo=datetime.timezone.utc, microsecond=0).isoformat()
        self['~timing'] = timing


class Answer(ThreadMixin, AriesProtocolMessage, metaclass=RegisterMessage):
    """Implementation of Answer

    https://github.com/hyperledger/aries-rfcs/tree/master/features/0113-question-answer
    """

    PROTOCOL = 'questionanswer'
    NAME = 'answer'

    def __init__(
            self, response: Optional[str] = None, thread_id: Optional[str] = None,
            out_time: Optional[str] = None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        if response is not None:
            self['response'] = response
        if thread_id is not None:
            self.thread_id = thread_id
        if out_time is not None:
            timing = self.get('~timing', {})
            timing['out_time'] = out_time
            self['~timing'] = timing

    @property
    def response(self) -> Optional[str]:
        return self.get('response', None)

    @property
    def out_time(self) -> Optional[str]:
        return self.get('~timing', {}).get('out_time', None)

    def set_out_time(self):
        timing = self.get('~timing', {})
        timing['out_time'] = datetime.datetime.now().astimezone().replace(microsecond=0).isoformat()
        self['~timing'] = timing

    async def sign(self, crypto: AbstractCrypto, q: Question, verkey: str):
        data = q.question_text or '' + self.response or '' + q.nonce or ''
        hashfunc = hashlib.sha256
        hasher = hashfunc()
        hasher.update(data.encode('utf-8'))
        digest = hasher.digest()
        self['response~sig'] = await sign(crypto, digest, verkey)
